;; (load "../manager/load")
;; (manage 'new 'experimental-layers)

;;; layer
;; todo: sanity checks
(define (layer name dependent-layers compute-default-value)
  (list 'layer name dependent-layers compute-default-value))
(define (layer-name layer)
  (cadr layer))
(define (layer-dependent-layers layer)
  (caddr layer))
(define (layer-compute-default-value layer)
  (cadddr layer))

(define base-layer
  (layer 'base
         '()
         (lambda (x)
           x)))

;;; layered data
;; todo: sanity checks
(define (layered-datum base-value layer-bindings)
  (flatten-layers
   (make-entity
    dispatch-layers
    (cons* 'layered-datum base-value layer-bindings))))
(define (layered-datum-base-value datum)
  (cadr (entity-extra datum)))
(define (layered-datum-layer-bindings datum)
  (cddr (entity-extra datum)))
(define (layered-datum? obj)
  (and (entity? obj)
       (pair? (entity-extra obj))
       (eq? 'layered-datum (car (entity-extra obj)))))

;; register predicate so that we can print layered data nicely
;; note: mit-scheme's register-predicate! conflicts with sdf's...
;; (register-predicate! layered-datum? 'layered-datum)
;; hack to be compatible with sdf
((access register-predicate! system-global-environment)
 layered-datum? 'layered-datum)

;; make layered data print nicely
(define-print-method layered-datum?
  (standard-print-method
   'layers
   (lambda (datum)
     (cons (layered-datum-base-value datum)
           (map (lambda (layer-binding)
                  (list (layer-name (car layer-binding))
                        (cdr layer-binding)))
                (layered-datum-layer-bindings datum))))))

;; have a slightly nicer syntax
;; todo: make layers-syntax file
(define-syntax layers                   
  (syntax-rules ()
    ((_ base-value (layer value) ...)
     (layered-datum base-value (list (cons layer value) ...)))))

;;; flatten-layers
;; flatten-layers ensures that the base value is not a layered datum.
;; If the base value is a layered datum, then merge the layer
;; bindings
(define (flatten-layers datum)
  (let ((base-value (layered-datum-base-value datum)))
    (if (layered-datum? base-value)
        (layered-datum
         (layered-datum-base-value base-value)
         (merge-layer-bindings (layered-datum-layer-bindings datum)
                               (layered-datum-layer-bindings base-value)))
        datum)))

;; combine layer bindings, keeping those of bindings1 if both bindings
;; have them
;; idea: maybe each layer can have its own way of merging
(define (merge-layer-bindings bindings1 bindings2)
  (define (maybe-add-binding binding bindings)
    (if (bindings-have-layer? bindings (car binding))
        bindings
        (cons binding bindings)))
  (fold maybe-add-binding bindings1 bindings2))

;;; dispatch-layers
;; dispatch-layers applies a layered procedure on arguments
(define (dispatch-layers proc . args)
  (define (get-arguments-per-layer layer)
    (map (lambda (arg)
           (get-layer-value arg layer))
         args))
  (define (get-layer-value-per-layer layer)
    (apply (get-layer-value proc layer)
           (map get-arguments-per-layer
                (cons layer
                      (layer-dependent-layers layer)))))
  (let ((all-args-layers (apply lset-union
                                eq?
                                (map get-object-layers args)))
        (base-result (apply (layered-datum-base-value proc)
                            args)))
    (let ((layers-to-be-computed
           (lset-intersection eq?
                              all-args-layers
                              (get-object-layers proc))))
      (if (null? layers-to-be-computed)
          base-result
          (layered-datum
           base-result
           (map (lambda (layer)
                  (cons layer
                        (get-layer-value-per-layer layer)))
                layers-to-be-computed))))))

;; dispatch-layers applies the base value to the entire arguments, and
;; we flatten-layers the result during creation of layered
;; data. Because of how flatten-layers merges layers, we can easily
;; layer procedures which produce layered datums, while preserving the
;; original procedures' behavior.

;;; details (helper functions)
(define (default-value layer base-value)
  ((layer-compute-default-value layer) base-value)
  ;; omit internal projection; do we need it?
  )

;; if the layer exists in datum, then get the corresponding
;; value. Otherwise, get the default value for the layer and base
;; value
(define (get-layer-value datum layer)
  (cond ((layered-datum? datum)
         (let ((binding (assq layer
                              (layered-datum-layer-bindings datum))))
           (if binding
               (cdr binding)
               (default-value layer
                 (layered-datum-base-value datum)))))
        (else
         (default-value layer datum))))

(define (get-object-layers obj)
  (if (layered-datum? obj)
      (map car (layered-datum-layer-bindings obj))
      '()))

(define (bindings-have-layer? bindings layer)
  (assq layer bindings))

(define (get-base-value obj)
  (if (layered-datum? obj)
      (layered-datum-base-value obj)
      obj))

;; used to help write compute-default-value for layers
;; if base value is a procedure, then the layer value is a procedure
;; which returns default-value. Otherwise, the layer value is just the
;; default-value
(define ((simple-compute-default-value default-value) base-value)
  (if (procedure? base-value)
      ;; assumes base-value is not a higher-order procedure...
      (lambda args default-value)
      default-value))

;; the following is really hacky. todo: make it less hacky.
(define current-environment
  (the-environment))

(define (make-layered! proc-name)
  (let ((old-proc (environment-lookup current-environment proc-name)))
    (if (not (layered-datum? old-proc))
        (environment-define
         current-environment
         proc-name
         (layers (lambda args
                   (apply old-proc
                          (map get-base-value args))))))))

(define (install-handler! layer proc-name handler)
  (make-layered! proc-name)
  (environment-define
   current-environment
   proc-name
   (layers (environment-lookup current-environment proc-name)
           (layer handler))))

;;; lastly, fix if
;; if needs to access the base value of a layered datum, if it's
;; passed a layered datum. Note: somewhat dangerous!
(define (if-2 pred then-thunk)
  (if (get-base-value pred)
      (then-thunk)))
(define (if-3 pred then-thunk else-thunk)
  (if (get-base-value pred)
      (then-thunk)
      (else-thunk)))

(define-syntax if
  (syntax-rules ()
    ((_ pred then)
     (if-2 pred (lambda () then)))
    ((_ pred then else)
     (if-3 pred (lambda () then) (lambda () else)))))
